// This Software demo for the Precision Microdrives Haptic Shield is provided
// under the MIT License

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to 
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.


// Change these defines to change the library and effect used when the test is complete
#define	FINAL_LIBRARY	2
#define	FINAL_EFFECT	83

// Include system headers
#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include <SPI.h>
#include <Wire.h>

// Include our headers
#include "types.h"
#include "defs.h"
#include "Metro.h"
#include "DRV2605.h"
#include "SSD1306LessRAM.h"

// Libraries
#include "mpr121.h"
#include "i2c.h"

// Objects
DRV2605 drv2605 = DRV2605();
Metro timer = Metro( 100 );
SSD1306	oled = SSD1306( OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS );

static uint16_t old_touchstatus = 0;
static uint16_t readTouchStatus = 0;
static uint8_t keys_state_change = 0;

// State of the test
enum State
{
	NONE,
	STARTING,
	WAITING_WIRE,
	AUTOCAL_RUN,
	AUTOCAL_SUCCESS,
	AUTOCAL_FAIL,
	CALIBRATING_M1,
	CALIBRATING_M2,
	CALIBRATING_M3,
	CALIBRATING_M4,
	FINISHED
};

State state = STARTING;

// Input
#define NUM_KEYS	6

typedef struct {
	uint8_t	key;
	bool	down;
	long	lastDown, lastUp;
} KeyInfo;

KeyInfo keys[ NUM_KEYS ] = {
	{ LEFT,			false,	0, 0 },
	{ SELECT,		false,	0, 0 },
	{ RIGHT,		false,	0, 0 },
	{ BACK,			false,	0, 0 },
	{ PLAY,			false,	0, 0 },
	{ PMD_BUTTON,	false,	0, 0 },
};

// Motor
typedef struct {
	uint8_t	id;
	boolean LRA;			// TRUE if LRA, else ERM.
	uint8_t min_duty;		// Minimum PWM duty cycle to start ERM 0 = 0%, 511 = 100%
	uint8_t rated_duty;		// Rated PWM duty cycle to run ERM 0 = 0%, 511 = 100% = 3.3V
	uint8_t max_duty;		// Maximum PWM duty cycle to run ERM 0 = 0%, 511 = 100%
	char part_num[8];		// Part number string.
} MotorInfo;

// The motor table
static const MotorInfo motorTable[ NUM_MOTORS ] PROGMEM ={
	{ 0,	false,	56,		66,		87,		"305-000" },	// M0 - ERM 305-000, min 1.1V, rated 1.3V, max 1.7V
	{ 1,	false,	51,		153,	184,	"306-109" },	// M1 - ERM 306-109, min 1.0V Rated 3.0V max 3.6V
	{ 2,	false,	77,		153,	153,	"308-102" },	// M2 - ERM 308-102, min 1.5V, rated 3.0V, max 3.0V
	{ 3,	true,	39,		110,	158,	"C10-100" },	// M3 - Leaded, LRA, C13-000 min 0.50 Vrms rated 2.30 Vrms max 3.30 Vrms
};

MotorInfo keyMotor, currentMotor;
uint8_t currentMotorId = 0;

void readMotorInfo( uint8_t id, MotorInfo* info );

uint8_t ac18, ac19, ac1A;

uint8_t motorCalibrations[ NUM_MOTORS ][ 3 ];

// Wire tests
#define NUM_DEVICES	3
byte addresses[ NUM_DEVICES ] = { 0x58, 0x5A, 0x5D };

// Strings
//																	= "1234567890132456789012"
// - Checking Wire
static const char MSG_CHECK_WIRE_1[] PROGMEM = "Checking I2C";
// - Wire failed
static const char MSG_WIRE_FAILED_1[] PROGMEM = "I2C check failed";
static const char MSG_WIRE_FAILED_2[] PROGMEM = "Touch failed: %#.2X";
static const char MSG_WIRE_FAILED_3[] PROGMEM = "DRV failed:   %#.2X";
const char * const DEVICE_MSG[ NUM_DEVICES ] PROGMEM = {
	&MSG_WIRE_FAILED_3[ 0 ],	// 0x58: DRV
	&MSG_WIRE_FAILED_3[ 0 ],	// 0x5A: DRV
	&MSG_WIRE_FAILED_2[ 0 ] 	// 0x5D: Touch
};
// - Running AutoCal
static const char MSG_RUNNING_AUTOCAL_1[] PROGMEM = "Running AutoCal";
static const char MSG_RUNNING_AUTOCAL_2[] PROGMEM = "Shield should give";
static const char MSG_RUNNING_AUTOCAL_3[] PROGMEM = "1 second of feedback";
// - AutoCal succeeded
static const char MSG_AUTOCAL_SUCCESS_1[] PROGMEM = "%s calibrated";
static const char MSG_AUTOCAL_SUCCESS_2[] PROGMEM = "Results are:";
static const char MSG_AUTOCAL_SUCCESS_3[] PROGMEM = "%#.2x %#.2x %#.2x";
static const char MSG_AUTOCAL_SUCCESS_4[] PROGMEM = "Calibrating %s";
static const char MSG_AUTOCAL_SUCCESS_5[] PROGMEM = "Cycle with PMD button";
// - AutoCal failed
static const char MSG_AUTOCAL_FAILED_1[] PROGMEM = "AutoCal failed.";
static const char MSG_AUTOCAL_FAILED_2[] PROGMEM = "Check serial output";
// - Finished
static const char MSG_FINISHED_1[] PROGMEM = "Playing on %s";
static const char MSG_FINISHED_2[] PROGMEM = "Using auto cals";

char lineBuf[ 22 ];

void setup()
{
	// Setup serial
	Serial.begin( 9600 );
	Serial.print( F("FreeMem=") );
	Serial.println( freeRAM() );
	
	// Begine Wire to test that all devices are connected correctly
	Wire.begin();
	
	// Setup pins
	pinMode( SW_LRA_M,		OUTPUT );	// Output for switching LRA+/- v.s. M+/-
	pinMode( SW_MOS_DRV,	OUTPUT );	// Output for switching +3.3V/MOS- v.s. DRV+/-
	pinMode( DRV_2605_EN,	OUTPUT );	// Output for ERV2605 ENable
	pinMode( PWM_OUT,		OUTPUT );	// Output for PWM

	digitalWrite( SW_LRA_M,		LOW );	// Select M+ and M- 
	digitalWrite( SW_MOS_DRV,	HIGH );	// Select +3.3V and MOS- 
	digitalWrite( DRV_2605_EN,	LOW );	// Disable the DRV2605 (low power mode)
	digitalWrite( PWM_OUT,		LOW );	// PWM output low for now 

	pinMode( MOTOR_PIN_0,	OUTPUT );	// Motor select 1
	pinMode( MOTOR_PIN_1,	OUTPUT );	// Motor select 2

	setupTouchPad();
	
	drv2605.init();
	oled.ssd1306_init( SSD1306_SWITCHCAPVCC );
	oled.clear_display();
	
	// Set up key motor
	readMotorInfo( 3, &keyMotor );
}

void loop()
{
	keys_service();
	
	byte a, b, c;
	
	switch( state )
	{
		case NONE:
			break;
			
		case STARTING:
			Serial.println( F("Doing STARTING state") );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_CHECK_WIRE_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			
			state = WAITING_WIRE;
			break;
			
		case WAITING_WIRE:
			Serial.println( F("Doing WAITING_WIRE state") );
			
			// Start with the failure state on the screen. 
			// This will be overwritten is we pass the test
			oled.clear_display();
			sprintf_P( lineBuf, MSG_WIRE_FAILED_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			
			// Code based off the i2c_scanner sample.
			// Uses Wire.endTransmission to see if a 
			// device is connected to an address.
			a = c = 0;
			for( b = 0 ; b < NUM_DEVICES ; ++b )
			{
				Wire.beginTransmission( addresses[ b ] );
				
				if( 0 != Wire.endTransmission() )
				{
					// Failed to reach a device
					a = 1;
					const char* str = (char*)pgm_read_word( &(DEVICE_MSG[ b ]) );
					sprintf_P( lineBuf, str, addresses[ b ] );
				
					oled.clear_buf();
					oled.drawstring( 0, 3 + c++, lineBuf, WHITE );
				}
			}
			
			Serial.print( F("I2C device check ") );
			if( 0 != a )
			{
				Serial.println( F("failed") );
				state = NONE;
			}
			else
			{
				Serial.println( F("passed") );
				sprintf_P( lineBuf, MSG_RUNNING_AUTOCAL_1 );
				oled.clear_buf();
				oled.drawstring( 0, 2, lineBuf, WHITE );
				sprintf_P( lineBuf, MSG_RUNNING_AUTOCAL_2 );
				oled.clear_buf();
				oled.drawstring( 0, 3, lineBuf, WHITE );
				sprintf_P( lineBuf, MSG_RUNNING_AUTOCAL_3 );
				oled.clear_buf();
				oled.drawstring( 0, 4, lineBuf, WHITE );
				
				state = AUTOCAL_RUN;
			}
			
			break;
			
		case AUTOCAL_RUN:
			Serial.println( F("Doing AUTOCAL_RUN state") );
			
			digitalWrite( SW_LRA_M,		HIGH );	// Select the LRA on the shield
			digitalWrite( SW_MOS_DRV,	LOW );	// Select DRV+ and DRV- 
			digitalWrite( DRV_2605_EN,	HIGH );	// Enable the DRV2605
			
			ac18 = ac19 = ac1A = 0;
			b = drv2605.autoCal( keyMotor.rated_duty, keyMotor.max_duty, keyMotor.LRA, &ac18, &ac19, &ac1A );
			if( b )
			{
				state = AUTOCAL_SUCCESS;
			}
			else
			{
				state = AUTOCAL_FAIL;
			}
			
			digitalWrite( DRV_2605_EN,	LOW );	// Disable the DRV2605 (low power mode)
			digitalWrite( SW_MOS_DRV,	HIGH );	// Select +3.3V and MOS- 
			digitalWrite( SW_LRA_M,		LOW );	// Select M+ and M-
			
			break;
			
		case AUTOCAL_SUCCESS:
			Serial.println( F("Doing AUTOCAL_SUCCESS state good") );
			readMotorInfo( currentMotorId, &currentMotor );
			
			oled.clear_display();
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_1, "Key motor" );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_3, ac18, ac19, ac1A );
			oled.clear_buf();
			oled.drawstring( 0, 4, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_4, currentMotor.part_num );
			oled.clear_buf();
			oled.drawstring( 0, 5, lineBuf, WHITE );
			
			state = CALIBRATING_M1;
			break;
			
		case AUTOCAL_FAIL:
			Serial.println( F("Doing AUTOCAL_FAIL state bad") );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_AUTOCAL_FAILED_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_FAILED_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = NONE;
			break;
			
		case CALIBRATING_M1:
			Serial.println( F("Doing CALIBRATING_M1 state") );
			if( calibrateGripMotor( 0 ) ) 
				state = CALIBRATING_M2;
			else
				state = AUTOCAL_FAIL;
			break;
			
		case CALIBRATING_M2:
			Serial.println( F("Doing CALIBRATING_M2 state") );
			if( calibrateGripMotor( 1 ) )
				state = CALIBRATING_M3;
			else
				state = AUTOCAL_FAIL;
			break;
			
		case CALIBRATING_M3:
			Serial.println( F("Doing CALIBRATING_M3 state") );
			if( calibrateGripMotor( 2 ) )
				state = CALIBRATING_M4;
			else
				state = AUTOCAL_FAIL;
			break;
			
		case CALIBRATING_M4:
			Serial.println( F("Doing CALIBRATING_M4 state") );
			if( calibrateGripMotor( 3 ) )
			{
				sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_5, currentMotor.part_num );
				oled.clear_buf();
				oled.drawstring( 0, 5, lineBuf, WHITE );
				
				state = FINISHED;
			}
			else
				state = AUTOCAL_FAIL;
			break;

		case FINISHED:
			digitalWrite( SW_LRA_M,		LOW );	// Select the grip motor
			digitalWrite( SW_MOS_DRV,	LOW );	// Select DRV+ and DRV- 
			digitalWrite( DRV_2605_EN,	HIGH );	// Enable the DRV2605
			
			drv2605.playFullHaptic( 
				currentMotor.LRA ? 6 : FINAL_LIBRARY, FINAL_EFFECT, 
				currentMotor.rated_duty, currentMotor.max_duty, 
				motorCalibrations[ currentMotorId ][ 0 ], motorCalibrations[ currentMotorId ][ 1 ], motorCalibrations[ currentMotorId ][ 2 ] );
			
			digitalWrite( DRV_2605_EN,	LOW );	// Disable the DRV2605 (low power mode)
			digitalWrite( SW_MOS_DRV,	HIGH );	// Select +3.3V and MOS- 
			digitalWrite( SW_LRA_M,		LOW );	// Select M+ and M-
			break;
	}
} // loop

boolean calibrateGripMotor( uint8_t motorId )
{
	digitalWrite( SW_LRA_M,		LOW );	// Select the grip motor
	digitalWrite( SW_MOS_DRV,	LOW );	// Select DRV+ and DRV- 
	digitalWrite( DRV_2605_EN,	HIGH );	// Enable the DRV2605
	
	ac18 = ac19 = ac1A = 0;
	
	currentMotorId = motorId;
	readMotorInfo( currentMotorId, &currentMotor );
	selectGripMotor( currentMotorId );
	
	boolean done = false;
	
	if( drv2605.autoCal( currentMotor.rated_duty, currentMotor.max_duty, currentMotor.LRA, &ac18, &ac19, &ac1A ) )
	{
		oled.clear_display();
		sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_1, currentMotor.part_num );
		oled.clear_buf();
		oled.drawstring( 0, 2, lineBuf, WHITE );
		sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_2 );
		oled.clear_buf();
		oled.drawstring( 0, 3, lineBuf, WHITE );
		sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_3, ac18, ac19, ac1A );
		oled.clear_buf();
		oled.drawstring( 0, 4, lineBuf, WHITE );
		sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_4, currentMotor.part_num );
		oled.clear_buf();
		oled.drawstring( 0, 5, lineBuf, WHITE );
		
		motorCalibrations[ motorId ][ 0 ] = ac18;
		motorCalibrations[ motorId ][ 1 ] = ac19;
		motorCalibrations[ motorId ][ 2 ] = ac1A;
		
		done = true;
	}
	else
	{
		state = AUTOCAL_FAIL;
	}
	
	digitalWrite( DRV_2605_EN,	LOW );	// Disable the DRV2605 (low power mode)
	digitalWrite( SW_MOS_DRV,	HIGH );	// Select +3.3V and MOS- 
	digitalWrite( SW_LRA_M,		LOW );	// Select M+ and M-
	
	return done;
}

void readMotorInfo( uint8_t id, MotorInfo* info )
{
	info->id = pgm_read_byte( &motorTable[ id ].id );
	info->LRA = pgm_read_byte( &motorTable[ id ].LRA );
	info->min_duty = pgm_read_byte( &motorTable[ id ].min_duty );
	info->rated_duty = pgm_read_byte( &motorTable[ id ].rated_duty );
	info->max_duty = pgm_read_byte( &motorTable[ id ].max_duty );
	info->LRA = pgm_read_byte( &motorTable[ id ].LRA );
	
	for( uint8_t i = 0 ; i < 7 ; ++i ) 
	{ // strncpy_P() doesn't like a volatile target, so do it like this:
		info->part_num[ i ] = pgm_read_byte( &motorTable[ id ].part_num[ i ] );
	}
	
	info->part_num[ 7 ] = 0; // Be sure zero terminated.
} // readMotorInfo

void selectGripMotor( uint8_t id )
{
	digitalWrite( MOTOR_PIN_0,	id % 2 == 0 ? LOW : HIGH );
	digitalWrite( MOTOR_PIN_1,	id > 1 ? HIGH : LOW );
} // selectGripMotor

void setupTouchPad()
{
	//make sure the interrupt pin is an input and pulled high
	pinMode( IRQ_PIN, INPUT );
	digitalWrite( IRQ_PIN, HIGH );
	
	//output on ADC4 (PC4, SDA)
	DDRC |= 0b00010000;
	// Pull-ups on I2C Bus
	PORTC = 0b00110000; 
	// initalize I2C bus. Wiring lib not used. 
	i2cInit( 200 );

	timer.reset();
	while( 0 == timer.check() );
	while( 0 == timer.check() ); // wait at least 1/10 sec.
	
	// initialize mpr121
	mpr121QuickConfig();
	
	// Create and interrupt to trigger when a button
	// is hit, the IRQ pin goes low, and the function getKey is run. 
	attachInterrupt( 0, keyPressed, FALLING );
} // setupTouchPad

// this is interrupt INT0
void keyPressed()
{
	keys_state_change = 1;
	readTouchStatus = mpr121Read( 0x01 ) << 8;
	readTouchStatus |= mpr121Read( 0x00 );
	detachInterrupt( 0 );
} // keyPressed

void keys_service()
{
	volatile uint16_t touchstatus;
	
	if( keys_state_change ) 
	{
		touchstatus = readTouchStatus;

		if( touchstatus & OVCF )
		{
			Serial.println( F("OVER VOLT. !!!" ));
			mpr121Write( 0x80, 0x63 ); // Perform soft reset
			mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
			mpr121Write( 0x80, 0x63 ); // Perform soft reset
			mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
			mpr121Write( 0x80, 0x63 ); // Perform soft reset
			mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
			mpr121QuickConfig(); // re-init mpr121
		}

		if( touchstatus != old_touchstatus )
		{
			uint8_t i = 0;
			uint8_t passed = 0;
			for( ; i < NUM_KEYS ; ++i )
			{
				if( keys[ i ].down && !(touchstatus & keys[ i ].key ) )
				{
					// Key up
					keys[ i ].down = false;
					keys[ i ].lastUp = millis();
					passed = 1;
					break;
				}
				else if( !keys[ i ].down && (touchstatus & keys[ i ].key ) )
				{
					// Key down
					keys[ i ].down = true;
					keys[ i ].lastDown = millis();
					passed = 1;
					
					if( keys[ i ].key == PMD_BUTTON )
					{
						currentMotorId = (currentMotorId + 1) % NUM_MOTORS;
						readMotorInfo( currentMotorId, &currentMotor );
						selectGripMotor( currentMotorId );
						
						oled.clear_display();
						sprintf_P( lineBuf, MSG_FINISHED_1, currentMotor.part_num );
						oled.clear_buf();
						oled.drawstring( 0, 2, lineBuf, WHITE );
						sprintf_P( lineBuf, MSG_FINISHED_2 );
						oled.clear_buf();
						oled.drawstring( 0, 3, lineBuf, WHITE );
						sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_3, motorCalibrations[ currentMotorId ][ 0 ], motorCalibrations[ currentMotorId ][ 1 ], motorCalibrations[ currentMotorId ][ 2 ] );
						oled.clear_buf();
						oled.drawstring( 0, 4, lineBuf, WHITE );
						sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_5 );
						oled.clear_buf();
						oled.drawstring( 0, 5, lineBuf, WHITE );
					}
					
					break;
				}
			}

			if( !passed && (keys[ i ].key == PLAY) )
			{
				keys[ i ].down = false;
				keys[ i ].lastUp = millis();
				touchstatus &= ~keys[ i ].key;
			}

			old_touchstatus = touchstatus;
		}
		
		keys_state_change = 0;
		attachInterrupt( 0, keyPressed, FALLING );
	}
} // keys_service

int freeRAM() 
{
	extern int __heap_start, *__brkval; 
	int v; 
	return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
} // freeRAM
