// This Software demo for the Precision Microdrives Haptic Shield is provided
// under the MIT License

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to 
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

// Include system headers
#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>

#include <EEPROM.h>
#include <SPI.h>

// Debug
#include "debug.h"

// Include our headers
#include "types.h"
#include "defs.h"
#include "Metro.h"
#include "menu.h"
#include "motor.h"

// Libraries
#include "mpr121.h"
#include "i2c.h"

Menu menu = Menu();
Metro timer = Metro( 100 );

KeyInfo keys[ NUM_KEYS ] = {
	{ LEFT,		false,	0, 0 },
	{ SELECT,	false,	0, 0 },
	{ RIGHT,	false,	0, 0 },
	{ BACK,		false,	0, 0 },
	{ PLAY,		false,	0, 0 },
	{ PMD_BUTTON,	false,	0, 0 },
};

uint16_t old_touchstatus = 0;
static uint8_t lock_keys = 0;      // store code of key which is pressed and not released
static uint8_t keys_state_change = 0;
volatile uint16_t touchstatus;

void setup()
{
	// Setup serial
	Serial.begin( 9600 );

#ifdef _DEBUG
	Serial.print(F("FreeMem=") );
        Serial.println( freeRAM() );
#endif // _DEBUG

	// Setup pins
	pinMode( SW_LRA_M,	OUTPUT );	// Output for switching LRA+/- v.s. M+/-
	pinMode( SW_MOS_DRV,	OUTPUT );	// Output for switching +3.3V/MOS- v.s. DRV+/-
	pinMode( DRV_2605_EN,	OUTPUT );	// Output for ERV2605 ENable
	pinMode( PWM_OUT,	OUTPUT );	// Output for PWM

	digitalWrite( SW_LRA_M,		LOW );	// Select M+ and M- 
	digitalWrite( SW_MOS_DRV,	HIGH );	// Select +3.3V and MOS- 
	digitalWrite( DRV_2605_EN,	LOW );	// Disable the DRV2605 (low power mode)
	digitalWrite( PWM_OUT,		LOW );	// PWM output low for now 

	pinMode( MOTOR_PIN_0,	OUTPUT );	// Motor select 1
	pinMode( MOTOR_PIN_1,	OUTPUT );	// Motor select 2

	setupTouchPad();

	// Init controllers
	timer.interval( 100 );

	menu.init();
	motor.init( KEY_MOTOR_ID );
	
	// Ensure all motors are calibrated
	for( uint8_t motorId = 0 ; motorId < 4 ; motorId++ )
	{
		motor.selectMotor( motorId );
	}

	// Start PMD logo animation
//	menu.playLogo( &timer );

	// Show first menu page
	menu.displayPage( 0 );
}

void loop()
{
       updateInput();  // calculate time for long play
       keys_service();  // moved from interrupt controlled by flag key_down_flag 
	   menu.update();
	// Check motor playing
}

//========================================================================================================

void setupTouchPad()
{
	//make sure the interrupt pin is an input and pulled high
	pinMode( IRQ_PIN, INPUT );
	digitalWrite( IRQ_PIN, HIGH );
	
	// only Pull-ups on I2C Bus SDA, SCL don't set it as output
	PORTC = 0b00110000; 
	// initalize I2C bus. Wiring lib not used. 
       
        i2cInit(200);    // f_SCL = 200kHz
       
	timer.reset();                // from metro
	while( 0 == timer.check() );
	while( 0 == timer.check() ); // wait at least 1/10 sec.
	
	// initialize mpr121
	mpr121QuickConfig();
	
        attachInterrupt( 0, keyPressed, FALLING );


} // setupTouchPad

//--------------------------------------------------------------------------------------------------------

void keys_service()
{
   if (keys_state_change){
    
      passInput( &keys[ keys_state_change - 1 ] );        // key_state_change is from 1 to 6 , pass input need argument 0 - 5
      keys_state_change = 0;          // 
   }   
}  

//========================================================================================================

// this is interrupt INT0 triggered always when one or more of keys is released or pressed
void keyPressed()
{
  
  touchstatus = mpr121Read_word( 0x00 );    // read two bytes from one transsmision (multiple bytes reading) - possible faster
    
    if( touchstatus & OVCF )
    {
            // Serial.println( F("OVER VOLT. !!!" ));
		mpr121Write( 0x80, 0x63 ); // Perform soft reset
		mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
		mpr121Write( 0x80, 0x63 ); // Perform soft reset
		mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
		mpr121Write( 0x80, 0x63 ); // Perform soft reset
		mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
		mpr121QuickConfig();       // re-init mpr121
    }
                   
                uint8_t i = 0;        
                for(; i < NUM_KEYS ; ++i )        // NUM_KEYS = 0 - 5
		{
			if(keys[ i ].down && !(touchstatus & keys[ i ].key ) )      // if lock_keys=0 check for down if key is down and touchstatus = 0 (all released) 
			{
				//Serial.print(F( "Key up 0x") );
				//Serial.println( keys[ i ].key, HEX );
				// Key up
				 if (lock_keys - 1 ==  i ){
                                    lock_keys = 0;                  // if key is released unlock others keys
                                    keys[ i ].down = false;
    				    keys[ i ].lastUp = millis();
    				    if (!keys_state_change) keys_state_change = i+1;
                                    //passInput( &keys[ i ] );        // take action when ley is short press
                                  }
				break;
			}
			else if(!lock_keys && !keys[ i ].down && (touchstatus & keys[ i ].key ))  // if no lock keys, check for down
			{
			        //Serial.print(F( "Key down 0x" ));
			        //Serial.println( keys[ i ].key, HEX );
				// Key down
				lock_keys = i + 1;//keys[ i ].key;             // lock others keys i=<0,NUM_KEYS> lock_keys = <1,NUM_KEYS-1>
                                keys[ i ].down = true;
				keys[ i ].lastDown = millis();
				if (!keys_state_change) keys_state_change = i+1;
                                //passInput( &keys[ i ] );        // reaction for short press when 
				break;
			}
		}//for i <num_keys    

  
} // keyPressed

//=================================================================================================================================================================================================================

void passInput( KeyInfo* keyInfo )
{
	if( motor.isPlaying() )  return;

	uint8_t playKeyPress = 0, selectMotor = 0;
	long interval;

	if( keyInfo->down )            // if current button is dowm
	{
		interval = keyInfo->lastDown - keyInfo->lastUp;    // calculate interval between lastDown and last up
	}
	else 
	{
		interval = keyInfo->lastUp - keyInfo->lastDown;    // calculate time from last up and last down
	}

	menu.input( keyInfo, interval );      // if PLAY - play haptics for short

} // passInput


//=========================================================================================================

// for long press if long press take action only for BACK and PMD for long press
// for long press if long press take action only for BACK and PMD for long press
void updateInput()
{
} // updateInput

