// This Software demo for the Precision Microdrives Haptic Shield is provided
// under the MIT License

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to 
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

// Include system headers
#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include <SPI.h>
#include <Wire.h>

// Include our headers
#include "types.h"
#include "defs.h"
#include "Metro.h"
#include "DRV2605.h"
#include "SSD1306LessRAM.h"

// Libraries
#include "mpr121.h"
#include "i2c.h"

// Objects
DRV2605 drv2605 = DRV2605();
Metro timer = Metro( 100 );
SSD1306	oled = SSD1306( OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS );

static uint16_t old_touchstatus = 0;
static uint8_t keys_state_change = 0;

// State of the test
enum State
{
	NONE,
	STARTING,
	WAITING_WIRE,
	AUTOCAL_RUN,
	AUTOCAL_SUCCESS,
	AUTOCAL_FAIL,
	WAITING_BACK,
	BUTTON_BACK,
	WAITING_PLAY,
	BUTTON_PLAY,
	WAITING_PMD,
	BUTTON_PMD,
	WAITING_LEFT,
	BUTTON_LEFT,
	WAITING_SELECT,
	BUTTON_SELECT,
	WAITING_RIGHT,
	BUTTON_RIGHT,
	FINISHED,
	
} ;

State state = STARTING;

// Input
#define NUM_KEYS	6

typedef struct {
	uint8_t	key;
	bool	down;
	long	lastDown, lastUp;
} KeyInfo;

KeyInfo keys[ NUM_KEYS ] = {
	{ LEFT,			false,	0, 0 },
	{ SELECT,		false,	0, 0 },
	{ RIGHT,		false,	0, 0 },
	{ BACK,			false,	0, 0 },
	{ PLAY,			false,	0, 0 },
	{ PMD_BUTTON,	false,	0, 0 },
};

State keyStates[ NUM_KEYS ] = {
	WAITING_LEFT,
	WAITING_SELECT,
	WAITING_RIGHT,
	WAITING_BACK,
	WAITING_PLAY,
	WAITING_PMD
};

State nextStates[ NUM_KEYS ] = {
	BUTTON_LEFT,
	BUTTON_SELECT,
	BUTTON_RIGHT,
	BUTTON_BACK,
	BUTTON_PLAY,
	BUTTON_PMD
};

uint8_t keyPressEffects[ NUM_KEYS ] = {
	13,	// Left effect -> Buzz 1
	7,	// Select effect -> Bump 1
	52,	// Right effect -> Pulse 1
	1,	// Back effect -> Strong Single Click 1
	27,	// Play effect -> Strong Double Click 1
	12,	// PMD effect -> Triple Click
};

// Motor
typedef struct {
  boolean LRA;			// TRUE if LRA, else ERM.
  uint8_t min_duty;		// Minimum PWM duty cycle to start ERM 0 = 0%, 511 = 100%
  uint8_t rated_duty;	// Rated PWM duty cycle to run ERM 0 = 0%, 511 = 100% = 3.3V
  uint8_t max_duty;		// Maximum PWM duty cycle to run ERM 0 = 0%, 511 = 100%
  char part_num[8];		// Part number string.
} MotorInfo;

MotorInfo motorInfo = { true,	39,		110,	158,	"C10-100" };	// M3 - Leaded, LRA, C13-000 min 0.50 Vrms rated 2.30 Vrms max 3.30 Vrms
uint8_t compensation, backEMF, feedback;

// Wire tests
#define NUM_DEVICES	3
byte addresses[ NUM_DEVICES ] = { 0x58, 0x5A, 0x5D };

// Strings
//																	= "1234567890132456789012"
// - Checking Wire
static const char MSG_CHECK_WIRE_1[] PROGMEM = "Checking I2C";
// - Wire failed
static const char MSG_WIRE_FAILED_1[] PROGMEM = "I2C check failed";
static const char MSG_WIRE_FAILED_2[] PROGMEM = "Touch failed: %#.2X";
static const char MSG_WIRE_FAILED_3[] PROGMEM = "DRV failed:   %#.2X";
const char * const DEVICE_MSG[ NUM_DEVICES ] PROGMEM = {
	&MSG_WIRE_FAILED_3[ 0 ],	// 0x58: DRV
	&MSG_WIRE_FAILED_3[ 0 ],	// 0x5A: DRV
	&MSG_WIRE_FAILED_2[ 0 ] 	// 0x5D: Touch
};
// - Running AutoCal
static const char MSG_RUNNING_AUTOCAL_1[] PROGMEM = "Running AutoCal";
static const char MSG_RUNNING_AUTOCAL_2[] PROGMEM = "Shield should give";
static const char MSG_RUNNING_AUTOCAL_3[] PROGMEM = "1 second of feedback";
// - AutoCal succeeded
static const char MSG_AUTOCAL_SUCCESS_1[] PROGMEM = "AutoCal passed.";
static const char MSG_AUTOCAL_SUCCESS_2[] PROGMEM = "Results are:";
static const char MSG_AUTOCAL_SUCCESS_3[] PROGMEM = "%#.2x %#.2x %#.2x";
static const char MSG_AUTOCAL_SUCCESS_4[] PROGMEM = "Press back button";
// - AutoCal failed
static const char MSG_AUTOCAL_FAILED_1[] PROGMEM = "AutoCal failed.";
static const char MSG_AUTOCAL_FAILED_2[] PROGMEM = "Check serial output";
// - Back button success
static const char MSG_BUTTON_BACK_SUCCESS_1[] PROGMEM = "Back button pressed";
static const char MSG_BUTTON_BACK_SUCCESS_2[] PROGMEM = "Press play button";
// - Play button success
static const char MSG_BUTTON_PLAY_SUCCESS_1[] PROGMEM = "Play button pressed";
static const char MSG_BUTTON_PLAY_SUCCESS_2[] PROGMEM = "Press PMD button";
// - PMD button success
static const char MSG_BUTTON_PMD_SUCCESS_1[] PROGMEM = "PMD button pressed";
static const char MSG_BUTTON_PMD_SUCCESS_2[] PROGMEM = "Press left button";
// - Left button success
static const char MSG_BUTTON_LEFT_SUCCESS_1[] PROGMEM = "Left button pressed";
static const char MSG_BUTTON_LEFT_SUCCESS_2[] PROGMEM = "Press select button";
// - Select button success
static const char MSG_BUTTON_SELECT_SUCCESS_1[] PROGMEM = "Select button pressed";
static const char MSG_BUTTON_SELECT_SUCCESS_2[] PROGMEM = "Press right button";
// - Right button success
static const char MSG_BUTTON_RIGHT_SUCCESS_1[] PROGMEM = "Test finished. Shield";
static const char MSG_BUTTON_RIGHT_SUCCESS_2[] PROGMEM = "should be playing";
static const char MSG_BUTTON_RIGHT_SUCCESS_3[] PROGMEM = "continuous ramp";
static const char MSG_BUTTON_RIGHT_SUCCESS_4[] PROGMEM = "effects";

char lineBuf[ 22 ];

void setup()
{
	// Setup serial
	Serial.begin( 9600 );
	Serial.print( "FreeMem=" );
	Serial.println( freeRAM() );
	
	// Begine Wire to test that all devices are connected correctly
	Wire.begin();
	
	// Setup pins
	pinMode( SW_LRA_M,		OUTPUT );	// Output for switching LRA+/- v.s. M+/-
	pinMode( SW_MOS_DRV,	OUTPUT );	// Output for switching +3.3V/MOS- v.s. DRV+/-
	pinMode( DRV_2605_EN,	OUTPUT );	// Output for ERV2605 ENable
	pinMode( PWM_OUT,		OUTPUT );	// Output for PWM

	digitalWrite( SW_LRA_M,		LOW );	// Select M+ and M- 
	digitalWrite( SW_MOS_DRV,	HIGH );	// Select +3.3V and MOS- 
	digitalWrite( DRV_2605_EN,	LOW );	// Disable the DRV2605 (low power mode)
	digitalWrite( PWM_OUT,		LOW );	// PWM output low for now 

	pinMode( MOTOR_PIN_0,	OUTPUT );	// Motor select 1
	pinMode( MOTOR_PIN_1,	OUTPUT );	// Motor select 2

	setupTouchPad();
	
	digitalWrite( SW_LRA_M,		HIGH );	// Select the LRA on the shield
	digitalWrite( SW_MOS_DRV,	LOW );	// Select DRV+ and DRV- 
	digitalWrite( DRV_2605_EN,	HIGH );	// Enable the DRV2605
	
	drv2605.init();
	oled.ssd1306_init( SSD1306_SWITCHCAPVCC );
	oled.clear_display();
}

void loop()
{
	keys_service();
	
	byte a, b, c;
	
	switch( state )
	{
		case NONE:
			break;
			
		case STARTING:
			Serial.println( "Doing STARTING state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_CHECK_WIRE_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			
			state = WAITING_WIRE;
			break;
			
		case WAITING_WIRE:
			Serial.println( "Doing WAITING_WIRE state" );
			
			// Start with the failure state on the screen. 
			// This will be overwritten is we pass the test
			oled.clear_display();
			sprintf_P( lineBuf, MSG_WIRE_FAILED_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			
			// Code based off the i2c_scanner sample.
			// Uses Wire.endTransmission to see if a 
			// device is connected to an address.
			a = c = 0;
			for( b = 0 ; b < NUM_DEVICES ; ++b )
			{
				Wire.beginTransmission( addresses[ b ] );
				
				if( 0 != Wire.endTransmission() )
				{
					// Failed to reach a device
					a = 1;
					const char* str = (char*)pgm_read_word( &(DEVICE_MSG[ b ]) );
					sprintf_P( lineBuf, str, addresses[ b ] );
				
					oled.clear_buf();
					oled.drawstring( 0, 3 + c++, lineBuf, WHITE );
				}
			}
			
			Serial.print( "I2C device check " );
			if( 0 != a )
			{
				Serial.println( "failed" );
				state = NONE;
			}
			else
			{
				Serial.println( "passed" );
				sprintf_P( lineBuf, MSG_RUNNING_AUTOCAL_1 );
				oled.clear_buf();
				oled.drawstring( 0, 2, lineBuf, WHITE );
				sprintf_P( lineBuf, MSG_RUNNING_AUTOCAL_2 );
				oled.clear_buf();
				oled.drawstring( 0, 3, lineBuf, WHITE );
				sprintf_P( lineBuf, MSG_RUNNING_AUTOCAL_3 );
				oled.clear_buf();
				oled.drawstring( 0, 4, lineBuf, WHITE );
				
				state = AUTOCAL_RUN;
			}
			
			break;
			
		case AUTOCAL_RUN:
			Serial.println( "Doing AUTOCAL_RUN state" );
			compensation = backEMF = feedback = 0;
			b = drv2605.autoCal( motorInfo.rated_duty, motorInfo.max_duty, motorInfo.LRA, &compensation, &backEMF, &feedback );
			if( b )
			{
				state = AUTOCAL_SUCCESS;
			}
			else
			{
				state = AUTOCAL_FAIL;
			}
			break;
			
		case AUTOCAL_SUCCESS:
			Serial.println( "Doing AUTOCAL_SUCCESS state good" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_3, compensation, backEMF, feedback );
			oled.clear_buf();
			oled.drawstring( 0, 4, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_SUCCESS_4 );
			oled.clear_buf();
			oled.drawstring( 0, 5, lineBuf, WHITE );
			
			state = WAITING_BACK;
			break;
			
		case AUTOCAL_FAIL:
			Serial.println( "Doing AUTOCAL_FAIL state bad" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_AUTOCAL_FAILED_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_AUTOCAL_FAILED_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = NONE;
			break;
			
		case BUTTON_BACK:
			Serial.println( "Doing BUTTON_BACK state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_BUTTON_BACK_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_BACK_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = WAITING_PLAY;
			break;
			
		case BUTTON_PLAY:
			Serial.println( "Doing BUTTON_PLAY state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_BUTTON_PLAY_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_PLAY_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = WAITING_PMD;
			break;
			
		case BUTTON_PMD:
			Serial.println( "Doing BUTTON_PMD state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_BUTTON_PMD_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_PMD_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = WAITING_LEFT;
			break;
			
		case BUTTON_LEFT:
			Serial.println( "Doing BUTTON_LEFT state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_BUTTON_LEFT_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_LEFT_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = WAITING_SELECT;
			break;
			
		case BUTTON_SELECT:
			Serial.println( "Doing BUTTON_SELECT state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_BUTTON_SELECT_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_SELECT_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			
			state = WAITING_RIGHT;
			break;
			
		case BUTTON_RIGHT:
			Serial.println( "Doing BUTTON_RIGHT state" );
			oled.clear_display();
			sprintf_P( lineBuf, MSG_BUTTON_RIGHT_SUCCESS_1 );
			oled.clear_buf();
			oled.drawstring( 0, 2, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_RIGHT_SUCCESS_2 );
			oled.clear_buf();
			oled.drawstring( 0, 3, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_RIGHT_SUCCESS_3 );
			oled.clear_buf();
			oled.drawstring( 0, 4, lineBuf, WHITE );
			sprintf_P( lineBuf, MSG_BUTTON_RIGHT_SUCCESS_4 );
			oled.clear_buf();
			oled.drawstring( 0, 5, lineBuf, WHITE );
			
			state = FINISHED;
			break;
			
		case FINISHED:
			// Nothing to do here
			do
			{
				drv2605.playFullHaptic( 6, 82, motorInfo.rated_duty, motorInfo.max_duty, compensation, backEMF, feedback );
				delay( 10 );
				drv2605.playFullHaptic( 6, 70, motorInfo.rated_duty, motorInfo.max_duty, compensation, backEMF, feedback );
				delay( 10 );
			} while ( true );
			break;
	}
}

void setupTouchPad()
{
	//make sure the interrupt pin is an input and pulled high
	pinMode( IRQ_PIN, INPUT );
	digitalWrite( IRQ_PIN, HIGH );
	
	//output on ADC4 (PC4, SDA)
	DDRC |= 0b00010000;
	// Pull-ups on I2C Bus
	PORTC = 0b00110000; 
	// initalize I2C bus. Wiring lib not used. 
	i2cInit( 200 );

	timer.reset();
	while( 0 == timer.check() );
	while( 0 == timer.check() ); // wait at least 1/10 sec.
	
	// initialize mpr121
	mpr121QuickConfig();
	
	// Create and interrupt to trigger when a button
	// is hit, the IRQ pin goes low, and the function getKey is run. 
	attachInterrupt( 0, keyPressed, FALLING );
} // setupTouchPad

// this is interrupt INT0 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
void keyPressed()
{
  keys_state_change = 1;
  detachInterrupt( 0 );
} // keyPressed

void keys_service()
{
	volatile uint16_t touchstatus;
	
	if( keys_state_change ) 
	{
		touchstatus = mpr121Read( 0x01 ) << 8;
		touchstatus |= mpr121Read( 0x00 );

		if( touchstatus & OVCF )
		{
			Serial.println( F("OVER VOLT. !!!" ));
			mpr121Write( 0x80, 0x63 ); // Perform soft reset
			mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
			mpr121Write( 0x80, 0x63 ); // Perform soft reset
			mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
			mpr121Write( 0x80, 0x63 ); // Perform soft reset
			mpr121Write( 0x01, 0x80 ); // clear overcurrent flag
			mpr121QuickConfig(); // re-init mpr121
		}

		if( touchstatus != old_touchstatus )
		{
			uint8_t i = 0;
			uint8_t passed = 0;
			for( ; i < NUM_KEYS ; ++i )
			{
				if( keys[ i ].down && !(touchstatus & keys[ i ].key ) )
				{
					// Key up
					keys[ i ].down = false;
					keys[ i ].lastUp = millis();
					passed = 1;
					break;
				}
				else if( !keys[ i ].down && (touchstatus & keys[ i ].key ) )
				{
					// Key down
					keys[ i ].down = true;
					keys[ i ].lastDown = millis();
					passed = 1;
					
					if( keyStates[ i ] == state )
					{
						drv2605.playFullHaptic( 6, keyPressEffects[ i ], motorInfo.rated_duty, motorInfo.max_duty, compensation, backEMF, feedback );
						state = nextStates[ i ];
					}
					break;
				}
			}

			if( !passed && (keys[ i ].key == PLAY) )
			{
				keys[ i ].down = false;
				keys[ i ].lastUp = millis();
				touchstatus &= ~keys[ i ].key;
			}

			old_touchstatus = touchstatus;
		}
		
		keys_state_change = 0;
		attachInterrupt( 0, keyPressed, FALLING );
	}
} // keyPressed

int freeRAM() 
{
	extern int __heap_start, *__brkval; 
	int v; 
	return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
} // freeRAM
